"""DNS Authenticator for Edge-Center."""

import logging
from typing import Any
from typing import Callable
from typing import Optional

from certbot import errors
from certbot.plugins import dns_common
from certbot.plugins.dns_common import CredentialsConfiguration

from . import api_ecenter
from .api_ecenter import ECenterConflictException

logger = logging.getLogger(__name__)


class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for Edge-Center

    This Authenticator uses the Edge-Center DNS API to fulfill a dns-01 challenge.
    """

    _docs_url = 'https://apidocs.edgecenter.ru/dns#section/Authentication'
    description = ('Obtain certificates using a DNS TXT record (if you are using Edge-Center for '
                   'DNS).')
    ttl = 300

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.token = self.email = self.password = None
        self.credentials: Optional[CredentialsConfiguration] = None
        self.api_url = None
        self.auth_url = None
        self.dns_api_url = None

    @classmethod
    def add_parser_arguments(
            cls, add: Callable[..., None], default_propagation_seconds: int = 10
    ) -> None:
        super().add_parser_arguments(add, default_propagation_seconds)
        add('credentials', help='Edge-Center credentials INI file.')

    def more_info(self) -> str:
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using the Edge-Center API.'

    def _validate_credentials(self, credentials: CredentialsConfiguration) -> None:
        self.token = credentials.conf('apitoken')
        self.email = credentials.conf('email')
        self.password = credentials.conf('password')
        self.auth_url = credentials.conf('auth_url')
        self.dns_api_url = credentials.conf('dns_api_url')
        self.api_url = credentials.conf('api_url')

        if self.token:
            if self.email or self.password:
                raise errors.PluginError('{}: dns_ecenter_email and dns_ecenter_password are '
                                         'not needed when using an API Token'
                                         .format(credentials.confobj.filename))
        elif self.email or self.password:
            if not self.email:
                raise errors.PluginError('{}: dns_ecenter_email is required when using a Global '
                                         'API Key. (should be email address associated with '
                                         'Edge-Center account)'.format(credentials.confobj.filename))
            if not self.password:
                raise errors.PluginError('{}: dns_ecenter_password is required when using a '
                                         'Global API Key. (see {})'
                                         .format(credentials.confobj.filename, self._docs_url))
        else:
            raise errors.PluginError(
                '{}: Either dns_ecenter_apitoken (recommended), or '
                'dns_ecenter_email and dns_ecenter_password are required.'
                ' (see {})'.format(credentials.confobj.filename, self._docs_url)
            )

    def _setup_credentials(self) -> None:
        self.credentials = self._configure_credentials(
            'credentials',
            'Edge-Center credentials INI file',
            None,
            self._validate_credentials
        )

    def _perform(self, domain: str, validation_name: str, validation: str) -> None:
        self._get_client().add_txt_record(domain, validation_name, validation, self.ttl)

    def _cleanup(self, domain: str, validation_name: str, validation: str) -> None:
        self._get_client().del_txt_record(domain, validation_name)

    def _get_client(self) -> "_ECenterClient":
        if not self.credentials:  # pragma: no cover
            raise errors.Error("Plugin has not been prepared.")
        if self.token:
            return _ECenterClient(
                token=self.token,
                api_url=self.api_url,
                dns_api_url=self.dns_api_url,
                auth_url=self.auth_url
            )
        return _ECenterClient(
            login=self.email,
            password=self.password,
            api_url=self.api_url,
            dns_api_url=self.dns_api_url,
            auth_url=self.auth_url
        )


class _ECenterClient:
    """
    Edge-Center client.
    """

    record_type = 'TXT'

    def __init__(self, *args, **kwargs) -> None:
        self.gcore = api_ecenter.ECenterClient(*args, **kwargs)

    def add_txt_record(
            self, domain: str, record_name: str, record_content: str, record_ttl: int
    ) -> None:
        """
        Add a TXT record using the supplied information.

        :param str domain: The domain to use for verification.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        :param int record_ttl: The record TTL (number of seconds that the record may be cached).
        :raises certbot.errors.PluginError: if an error occurs communicating with the Edge-Center DNS API
        """
        domain = self._find_zone_name(domain=domain)
        try:
            self.gcore.record_create(
                domain, record_name, self.record_type, data=self._data_for_txt(record_ttl, [record_content]),
            )
        except ECenterConflictException:
            logger.debug('Record already present on zone. Try to update record content')
            exist_record_content = self.gcore.record_content(domain, record_name, self.record_type)
            if record_content not in exist_record_content:
                exist_record_content.append(record_content)
            self.gcore.record_update(
                domain,
                record_name,
                self.record_type,
                data=self._data_for_txt(record_ttl, exist_record_content),
            )
        logger.debug('Successfully added TXT record with record_name: %s', record_name)

    def del_txt_record(self, domain: str, record_name: str) -> None:
        """
        Delete a TXT record using the supplied information.

        :param str domain: The domain to use for verification.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        """
        try:
            domain = self._find_zone_name(domain)
            self.gcore.record_get(domain, record_name, self.record_type)
        except (api_ecenter.ECenterNotFoundException, api_ecenter.ECenterConflictException) as err:
            logger.debug('Encountered error finding zone_id during deletion: %s', err)
            return
        self.gcore.record_delete(domain, record_name, self.record_type)
        logger.debug('Successfully deleted TXT record.')

    @classmethod
    def _data_for_txt(cls, ttl, contents: list) -> dict:
        """Preparing data for TXT record."""
        return {'resource_records': [{'content': [content], 'enabled': True} for content in contents], 'ttl': ttl}

    def _find_zone_name(self, domain: str) -> str:
        """
        Find the zone_name for a given domain.

        Args:
            domain: The domain for which to find the zone_id.

        Returns:
            The zone_id, if found.
        """
        limit = 100
        domain_slit_list = '.'.join(domain.split('.')[-2:])
        zone_name_guesses = dns_common.base_domain_name_guesses(domain)[:-1]
        zones_raw = self.gcore.zones({'limit': limit, 'name': domain_slit_list})
        zones = [zona.get('name') for zona in zones_raw]

        for zone_name in zone_name_guesses:
            if zone_name in zones:
                logger.debug('Found zone_name: %s for domain: %s', zone_name, domain)
                return zone_name
        raise errors.PluginError(
            'Unable to determine zone name for {0} using zone names: '
            '{1}. Please confirm that the domain name has been '
            'entered correctly and is already associated with the '
            'supplied Edge-Center account.'.format(domain, zone_name_guesses)
        )



import pytest
import responses
from responses import matchers
import json

from certbot_dns_ecenter.api_ecenter import ECenterClient


@pytest.fixture
def record_payload():
    return {
        'domain': 'example.com',
        'record_name': '_acme-challenge.example.com',
        'record_content': '123456790',
        'record_ttl': 300
    }


@pytest.fixture
def rrset_exists_two_records():
    return '''\
{
    "resource_records": [
        {
            "content": [
                "coexisting content"
            ],
            "enabled": true
        },
        {
            "content": [
                "123456790"
            ],
            "enabled": true
        }
    ],
    "ttl": 300
}'''


def txt_data_expected1():
    return {'resource_records': [{'content': ['text'], 'enabled': True}], 'ttl': 300}


def txt_data_expected2():
    return {'resource_records': [{'content': ['text1'], 'enabled': True}, {'content': ['text2'], 'enabled': True}], 'ttl': 300}


@pytest.fixture
def mock_auth():
    # auth
    responses.add(
        responses.POST,
        f'{ECenterClient._auth_url}/auth/jwt/login',
        json={'access': 'token'},
        status=200
    )


@pytest.fixture
def mock_get_zone(record_payload):
    # check domain
    responses.add(
        responses.GET,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}',
        json={'name': record_payload["domain"]},
        status=200
    )

@pytest.fixture
def mock_get_zones(record_payload):
    domain_2_level = '.'.join(record_payload["domain"].split('.')[-2:])
    params = {'limit': '100', 'name': domain_2_level}
    responses.add(
        responses.GET,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}',
        match=[responses.matchers.query_param_matcher(params)],
        json={'zones': [{'name': record_payload["domain"]}]},
        status=200
    )

@pytest.fixture
def mock_del_record(record_payload):
    # del
    responses.add(
        responses.DELETE,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={}, status=200
    )


@pytest.fixture
def mock_get_record(record_payload):
    # check record
    responses.add(
        responses.GET,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={}, status=200
    )


@pytest.fixture
def mock_post_record(record_payload):
    # check record
    responses.add(
        responses.POST,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={},
        status=200
    )


@pytest.fixture
def mock_dns_api(record_payload, rrset_exists_two_records, mock_get_zones):
    responses.add(
        responses.POST,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={},
        status=409
    )
    responses.add(
        responses.GET,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json=json.loads(rrset_exists_two_records),
        status=200
    )
    responses.add(
        responses.PUT,
        f'{ECenterClient._dns_api_url}/{ECenterClient._root_zones}/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={},
        status=200
    )
    yield responses



"""
The `certbot_dns_ecenter` plugin automates the process of
completing a ``dns-01`` challenge (`acme.challenges.DNS01`) by creating, and
subsequently removing, TXT records using the G-Core DNS API.

.. note::
   The plugin is not installed by default. It can be installed by command below.

   .. code-block:: bash

      pip install certbot-dns-ecenter


Named Arguments
---------------

========================================  =====================================
``--dns-ecenter-credentials``                  Edge-Center credentials INI file. (Required)
``--dns-ecenter-propagation-seconds``          The number of seconds to wait for DNS
                                          to propagate before asking the ACME
                                          server to verify the DNS record. (Default: 10)
========================================  =====================================


Credentials
-----------

Use of this plugin requires a configuration file containing Edge-Center DNS API
credentials.
You can use Edge-Center API Token, obtained from your
`profile panel <https://accounts.edgecenter.ru/profile/api-tokens>`_
or
use Edge-Center Authentication credentials (email and password)
for `login <https://auth.edgecenter.ru/login/signin>`_ page.

Edge-Center API Token is recommended authentication option.

The token needed by Certbot for add temporary TXT record to zone what you need certificates for.

.. code-block:: ini
   :name: token
   :caption: Example `ecenter.ini` file using restricted API Token (recommended):

   # Edge-Center API token used by Certbot
   dns_ecenter_apitoken = 0123456789abcdef0123456789abcdef01234567

.. code-block:: ini
   :name: auth
   :caption: Example `gcore.ini` file using authentication credentials (not recommended):

   # Edge-Center API credentials used by Certbot
   dns_ecenter_email = ec_user@example.com
   dns_ecenter_password = 0123456789abcdef0123456789abcdef01234

The path to this file can be provided interactively or using the
``--dns-ecenter-credentials`` command-line argument. Certbot records the path
to this file for use during renewal, but does not store the file's contents.

.. caution::
   You should protect these API credentials as you would the password to your
   Edge-Center account. Users who can read this file can use these credentials
   to issue arbitrary API calls on your behalf. Users who can cause Certbot to
   run using these credentials can complete a ``dns-01`` challenge to acquire
   new certificates or revoke existing certificates for associated domains,
   even if those domains aren't being managed by this server.

Certbot will emit a warning if it detects that the credentials file can be
accessed by other users on your system. The warning reads "Unsafe permissions
on credentials configuration file", followed by the path to the credentials
file. This warning will be emitted each time Certbot uses the credentials file,
including for renewal, and cannot be silenced except by addressing the issue
(e.g., by using a command like ``chmod 600`` to restrict access to the file).

Also you can specify the Edge-Center ``auth`` and ``dns`` urls.

.. code-block:: ini
   :name: urls
   :caption: Example `ecenter.ini` file using alternative `auth` and `dns` urls.

    # Edge-Center API credentials used by Certbot
    dns_ecenter_auth_url = https://auth.example.com
    dns_ecenter_api_url = https://dns_api.example.com

Examples
--------

.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``

   certbot certonly \\
     --authenticator dns-ecenter \\
     --dns-ecenter-credentials=./ecenter.ini \\
     -d 'example.com'

.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``, waiting 60 seconds
             for DNS propagation

   certbot certonly \\
     --authenticator dns-ecenter \\
     --dns-ecenter-credentials=./ecenter.ini \\
     --dns-ecenter-propagation-seconds=60 \\
     -d 'example.com'
"""



import json
import pytest
import responses

from certbot_dns_ecenter.dns_ecenter import _ECenterClient
from tests.conftest import txt_data_expected1, txt_data_expected2


@pytest.mark.parametrize('kwargs', ({'token': None}, {'login': 'user'}, {'password': 'test'}))
def test_gcoreclient_fail(kwargs):
    # check
    with pytest.raises(ValueError):
        _ECenterClient(**kwargs)


@responses.activate
@pytest.mark.parametrize('kwargs', ({'token': '123'}, {'login': 'user', 'password': 'test'}))
def test_add_txt_record_success(kwargs, record_payload, mock_auth, mock_get_zones, mock_post_record):
    # act # check
    assert _ECenterClient(**kwargs).add_txt_record(**record_payload) is None


@responses.activate
def test_update_txt_record_no_duplicates(record_payload, mock_auth, mock_dns_api, rrset_exists_two_records, kwargs={'token': '123'}):
    _ECenterClient(**kwargs).add_txt_record(**record_payload)
    # post
    assert mock_dns_api.calls[1].request.body == b'{"resource_records": [{"content": ["123456790"], "enabled": true}], "ttl": 300}'
    # post: should get conflict on post
    assert mock_dns_api.calls[1].response.status_code == 409
    # get: should get 2 records
    assert mock_dns_api.calls[2].response._content == b'{"resource_records": [{"content": ["coexisting content"], "enabled": true}, {"content": ["123456790"], "enabled": true}], "ttl": 300}'
    # put: expect the original payload (no duplicate) in put request not to cause _two or more records have same content_ dnsapi error
    # could have checked count only, but comparing original content with payload
    assert sorted(json.loads(mock_dns_api.calls[3].request.body).items()) == sorted(json.loads(rrset_exists_two_records).items())


@responses.activate
@pytest.mark.parametrize('kwargs', ({'token': '123'}, {'login': 'user', 'password': 'test'}))
def test_del_txt_record_success(kwargs, record_payload, mock_auth, mock_get_zones, mock_get_record, mock_del_record):
    # init
    record_payload.pop('record_content')
    record_payload.pop('record_ttl')

    # act # check
    assert _ECenterClient(**kwargs).del_txt_record(**record_payload) is None


@pytest.mark.parametrize('in_data, out_data', (
    (['text'], txt_data_expected1()), (['text1', 'text2'], txt_data_expected2()),
))
def test_data_for_txt(in_data, out_data):
    # check
    assert _ECenterClient._data_for_txt(300, in_data) == out_data


@responses.activate
@pytest.mark.parametrize('subdomain', ('test1.example.com', 'test2.test1.example.com', 'test3.test2.test1.example.com'))
def test_find_zone_name_success(record_payload, subdomain, mock_auth, mock_get_zones):
    # check
    assert _ECenterClient(token='test')._find_zone_name(subdomain) == record_payload['domain']



"""Wrapper for Edge-Center DNS API."""

import http
import logging
import re
import typing
import urllib.parse

import requests
from requests import Session

logger = logging.getLogger(__name__)


class ECenterException(Exception):
    """Edge-Center DNS API client exception."""


class ECenterConflictException(Exception):
    """Edge-Center DNS API conflict exception."""


class ECenterNotFoundException(Exception):
    """Edge-Center DNS API conflict exception."""


class ECenterClient:
    """Edge-Center DNS API client."""

    _root_zones = 'v2/zones'
    _dns_api_url = 'https://api.edgecenter.ru/dns'
    _auth_url = 'https://api.edgecenter.ru/iam'
    _timeout = 10.0
    _error_format = 'Error %s. %s: %r, data: "%r", response: %s'

    def __init__(self, token=None, login=None, password=None, api_url=None, dns_api_url=None, auth_url=None):
        self._session = Session()
        if token is not None:
            self._session.headers.update({'Authorization': f'APIKey {token}'})
        elif login is not None and password is not None:
            token = self._auth(self._auth_url, login, password)
            self._session.headers.update({'Authorization': f'Bearer {token}'})
        else:
            raise ValueError('either token or login & password must be set')
        if api_url:
            self._auth_url = self._build_url(api_url, '/iam')
            self._dns_api_url = self._build_url(api_url, '/dns')
            # more specific parameters should win thus go latter
        if dns_api_url:
            self._dns_api_url = dns_api_url
        if auth_url:
            self._auth_url = auth_url

    def _auth(self, url, login, password):
        """Get auth token."""
        responce = self._session.request(
            'POST',
            self._build_url(url, 'auth', 'jwt', 'login'),
            json={'username': login, 'password': password},
        )
        responce.raise_for_status()
        return responce.json()['access']

    def _request(self, method: str, url: str, params=None, data=None) -> requests.Response or requests.RequestException:
        """Requests handler."""
        responce = self._session.request(method, url, params=params, json=data, timeout=self._timeout)
        if responce.status_code in (  # pylint: disable=R1720
                http.HTTPStatus.BAD_REQUEST, http.HTTPStatus.INTERNAL_SERVER_ERROR,
        ):
            logger.error(self._error_format, responce.status_code, method, url, data or params, responce.text)
            raise ECenterException(responce.text)
        elif responce.status_code == http.HTTPStatus.CONFLICT:
            raise ECenterConflictException(
                self._error_format % (responce.status_code, method, url, data or params, responce.text))
        elif responce.status_code == http.HTTPStatus.NOT_FOUND:
            raise ECenterNotFoundException(
                self._error_format % (responce.status_code, method, url, data or params, responce.text))
        responce.raise_for_status()
        return responce

    def zone(self, zone_name: str, params: dict = None) -> dict:
        """Get DNS zone."""
        return self._request('GET', self._build_url(self._dns_api_url, self._root_zones, zone_name), params).json()

    def zones(self, params: dict = None) -> dict:
        """Get DNS zones."""
        return self._request('GET', self._build_url(self._dns_api_url, self._root_zones), params).json()['zones']

    def zone_create(self, zone_name: str) -> dict:
        """Create DNS zone."""
        return self._request(
            'POST', self._build_url(self._dns_api_url, self._root_zones), data={'name': zone_name},
        ).json()

    def zone_records(self, zone_name: str) -> list:
        """List DNS records from zone."""
        url = self._build_url(self._dns_api_url, self._root_zones, zone_name, 'rrsets')
        rrsets = self._request('GET', url, params={'all': 'true'}).json()
        records = rrsets['rrsets']
        return records

    def record_create(self, zone_name: str, rrset_name: str, type_: str, data: dict) -> None:
        """Create DNS record in zone."""
        self._request('POST', self._rrset_url(zone_name, rrset_name, type_), data=data)

    def record_update(self, zone_name: str, rrset_name: str, type_: str, data: dict) -> None:
        """Update DNS record in zone."""
        self._request('PUT', self._rrset_url(zone_name, rrset_name, type_), data=data)

    def record_get(self, zone_name: str, rrset_name: str, type_: str) -> dict:
        """Get DNS record in zone."""
        return self._request('GET', self._rrset_url(zone_name, rrset_name, type_)).json()

    def record_content(self, zone_name: str, rrset_name: str, type_: str) -> typing.List[str]:
        """Get record content."""
        response = self.record_get(zone_name, rrset_name, type_)
        return [record['content'][0] for record in response['resource_records']]

    def record_delete(self, zone_name: str, rrset_name: str, type_: str) -> None:
        """Delete DNS record in zome."""
        self._request('DELETE', self._rrset_url(zone_name, rrset_name, type_))

    def _rrset_url(self, zone_name: str, rrset_name: str, type_: str) -> str:
        """Build full DNS URL for request."""
        return self._build_url(self._dns_api_url, self._root_zones, zone_name, rrset_name, type_)

    @staticmethod
    def _build_url(base: str, *items: typing.Iterable) -> typing.AnyStr:
        if not re.match(r'^https?://', base):
            raise ECenterException('Error schema url: please, check schema in url: "%s"' % base)
        for item in items:
            base = base.strip('/') + '/'
            base = urllib.parse.urljoin(base, item)
        return base



from setuptools import find_packages
from setuptools import setup
from certbot_dns_ecenter.__version__ import VERSION


install_requires = [
    'certbot>=1.23.0',
    'setuptools>=39.0.1',
]

docs_extras = [
    'Sphinx>=1.0',
    'sphinx_rtd_theme',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='certbot-dns-ecenter',
    version=VERSION,
    description="Edge-Center DNS Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Edge-Center/ec-dns-certbot-plugin',
    author="Edge-Center",
    author_email='support@edgecenter.ru',
    license='Apache License 2.0',
    python_requires='>=3.8',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'docs': docs_extras,
    },
    entry_points={
        'certbot.plugins': [
            'dns-ecenter = certbot_dns_ecenter.dns_ecenter:Authenticator',
        ],
    },
)
